//
//  SDBootLoader01.c
//
//  Program main / SDcard boot loader 0.1
//  Copyright (C) 2010 Toyohiko Togashi tog001@nifty.com
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		Self program writer from SDcard device.
//
//	Setting
//		AVR
//			Device	: ATMEL AVR ATmega328P
//			IDE		: AVR Studio 4.18.700  
//			Compiler: WinAVR 20100110(GCC 4.3.3)
//			Writer	: AVRISP mkII
//				Fuse bits: BOOTSZ=2048word, BOOTRST=Enable
//
//	Usage
//		Write .HEX file on SDcard. (FAT16 format)
//		Insert SDcard memory and mode switch on.
//		Turn on target circuit.
//		Status LED Blink->Writing, EverOn->error, OFF->OK reboot
//
//		Status LED:
//			Port	: Port D2 / Active low
//			
//		Mode switch:
//			Port	: Port B1 / Active low
//			
//		SDcard I/O:
//			SS		: Port B2
//			SCK 	: Port B3
//			MISO	: Port B4
//			MOSI	: Port B5
//
//		SDcard file:
//			File name	: "M328P.HEX" only root directory
//			Format		: INTEL .hex file format
//			
//
//
//	Update history
//	---------- ----- -------------------------------------------
//	2010.09.04 v0.1  First release
//	2010.12.20 v0.1a Bug fix, lost of last record
//
//

#include <avr/io.h>
#include <avr/boot.h>

#include <inttypes.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
    
#include "sd.h"


//
//	config
//
#define FILENAME		"M328P.HEX"

#define LED_1_INIT()	DDRD  |=  _BV(DDD2)
#define LED_1_ON()		PORTD &= ~_BV(PORTD2)
#define LED_1_OFF()		PORTD |=  _BV(PORTD2)
#define LED_1_REV()		PORTD ^=  _BV(PORTD2)

#define SWITCH_1_INIT()	DDRB  &= ~_BV(DDB1)		// CD(card detect signal) pin
#define SWITCH_1()		!(PINB & _BV(PINB1))
//#define SWITCH_1_INIT()	DDRD  &= ~_BV(DDD5); PORTD |= _BV(PORTD5)	//   enable internal pull-ups
//#define SWITCH_1()		!(PIND & _BV(PIND5))

#define RECTYPE_DATA	0x00			// HEX file format 
#define RECTYPE_END		0x01			//

#define PAGELIM			28672			// 28kB Maximum target program size


//
//	Flash programing	(boot.h sample code)
//
void boot_program_page(uint32_t page, uint8_t *buf) {
	uint16_t 	i;
	uint8_t 	sreg;

	// Disable interrupts.

	sreg = SREG;
	cli();

	eeprom_busy_wait();

	boot_page_erase(page);
	boot_spm_busy_wait ();      // Wait until the memory is erased.

	for (i=0; i<SPM_PAGESIZE; i+=2) {
	
	    // Set up little-endian word.

	    uint16_t w = *buf++;
	    w += (*buf++) << 8;

	    boot_page_fill (page + i, w);
	}

	boot_page_write (page);     // Store buffer in flash page.
	boot_spm_busy_wait();       // Wait until the memory is written.

	// Reenable RWW-section again. We need this if we want to jump back
	// to the application after bootloading.

	boot_rww_enable ();

	// Re-enable interrupts (if they were ever enabled).

	SREG = sreg;
}



//
//	M a i n
//
__attribute__ ((noreturn))				// code size reduction
void main(){

	// I/O config & init
	LED_1_INIT();
	LED_1_ON();
	SWITCH_1_INIT();

	// Request on program write
	if (SWITCH_1()) {
		int 			fd;
		unsigned short	now_page;

		// Open the SDcard
		if (sd_mount(0) < 0) {
			for(;;);
//			asm volatile ("jmp 0x0000");		// Normal booting
		}
		if ((fd = sd_open(FILENAME, 0, 0)) < 0) {
			asm volatile ("jmp 0x0000");		// Normal booting
		}

		// Read & write
		now_page = PAGELIM;
		for(;;){
			unsigned char	len;
			unsigned short	addr;
			unsigned char	type;
			char 			rec[80];
			char			*p;
			unsigned char 	bin[40];
			unsigned char 	*q;
			unsigned char 	mem[SPM_PAGESIZE];

			// Read record
			p = rec;
			while(sd_read(fd, p, 1) > 0) {
				if (*p == '\n') {
					break;
				}
				p++;
			}
			*p = '\0';

			// Hex to bin
			p = &rec[1];
			q = bin;
			while(*p > ' ') {
				unsigned char	i;
				unsigned char	b;

				b = 0;
				for(i = 0; i < 2; i++) {
					b <<= 4;
					if (*p < 'A') {
						b |= (*p++ - '0');
					} else {
						b |= (*p++ - 'A' + 10);
					}
				}
				*q++ = b;
			}

			// Build a data & write
			type = bin[3];
			switch(type) {
			case RECTYPE_END:
			case RECTYPE_DATA:
				len  =  bin[0];
				addr = (bin[1] << 8) | bin[2];
				q = &bin[4];
				for(; len > 0; --len) {
					unsigned short	page;
					unsigned char	offs;

					page = addr & ~(SPM_PAGESIZE - 1);
					offs = addr &  (SPM_PAGESIZE - 1);
					if (page != now_page) {
						if (now_page < PAGELIM) {
							boot_program_page(now_page, mem);
							LED_1_REV();
						}
						now_page = page;

					}
					mem[offs] = *q++;
					addr++;
				}
				break;
			default:
				break;
			}

			// Terminate
			if (type == RECTYPE_END) {
				if (now_page != PAGELIM) {				// ADD 2010.12.20
					boot_program_page(now_page, mem);	// ADD 2010.12.20
				}										// ADD 2010.12.20
				break;
			}
		}
		LED_1_OFF();
	}

	asm volatile ("jmp 0x0000");		// Normal booting
	exit(1);
}
